@@ -27,13 +27,13 @@ from gi.repository import Gtk, Gdk, GLib, Gio
 from datetime import date as dt_date, time as dt_time, datetime as dt_datetime, timedelta
 from dateutil import rrule as du_rrule
 from dateutil.relativedelta import relativedelta
-from icalendar import Event as iEvent
+from icalendar import Calendar as iCalendar, Event as iEvent, Todo as iTodo
 from importlib import import_module
 from os import path as ospath
 from sys import stderr
 import signal
 import ctypes
-from typing import Optional, Tuple
+from typing import Optional, Tuple, List
 
 # for internationalisation/localisation
 import locale
@@ -55,7 +55,7 @@ class GUI:
     _CSS_FILE_APP = '{:s}/css/pygenda.css'.format(ospath.dirname(__file__))
     _CSS_FILE_USER = GLib.get_user_config_dir() + '/pygenda/pygenda.css'
     _LOCALE_DIR = '{:s}/locale/'.format(ospath.dirname(__file__))
-    _VIEWS = ('Week','Year') # Order gives order in menus
+    _VIEWS = ('Week','Year','Todo') # Order gives order in menus
     SPINBUTTON_INC_KEY = (Gdk.KEY_plus,Gdk.KEY_greater)
     SPINBUTTON_DEC_KEY = (Gdk.KEY_minus,Gdk.KEY_less)
     STYLE_ERR = 'dialog_error'
@@ -65,6 +65,29 @@ class GUI:
     today_toggle_date = None
     today_toggle_idx = 0
 
+    views = []
+    view_widgets = [] # type: List[Gtk.Widget]
+    _view_idx = 0 # take first view as default
+
+    _lib_clip = None
+
+    _builder = Gtk.Builder()
+    _window = None # type: Gtk.Window
+    _is_fullscreen = False
+    _box_view_cont = None # type: Gtk.Box
+    _eventbox = Gtk.EventBox()
+
+    date_order = ''
+    date_formatting_numeric = ''
+    date_formatting_text = ''
+    date_formatting_text_noyear = ''
+    date_formatting_textabb = ''
+    date_formatting_textabb_noyear = ''
+
+    # For startup
+    _starting_cal = True
+    _loading_indicator = None
+
     Config.set_defaults('global',{
         'language': '', # use OS language
         'hide_titlebar_when_maximized': False,
@@ -96,7 +119,6 @@ class GUI:
         cls._init_locale()
 
         # Construct GUI from GTK Builder XML glade file
-        cls._builder = Gtk.Builder()
         cls._builder.add_from_file(cls._GLADE_FILE)
 
         cls._window = cls._builder.get_object('window_main')
@@ -123,6 +145,7 @@ class GUI:
             'menuitem_copy': cls.copy_request,
             'menuitem_paste': cls.paste_request,
             'menuitem_newentry': cls.event_newentry,
+            'menuitem_newtodo': cls.event_newtodo,
             'menuitem_edittime': cls.event_edittime,
             'menuitem_editrepeats': cls.event_editrepeats,
             'menuitem_editalarm': cls.event_editalarm,
@@ -213,7 +236,6 @@ class GUI:
         # First, initialise calendar connector.
         # Do this outside main thread, so UI remains active/responsive.
         ci_cancel = Gio.Cancellable.new()
-        cls._starting_cal = True
         if 'run_in_thread' in dir(Gio.Task):
             # Preferred way to run in separate thread.
             task = Gio.Task.new(cancellable=ci_cancel, callback=cls.init_cal_callback)
@@ -226,6 +248,7 @@ class GUI:
         cls._init_clipboard()
         cls._init_date_format()
         EntryDialogController.init()
+        TodoDialogController.init()
 
         # If date set from command line, jump there now
         if Config.date:
@@ -253,14 +276,12 @@ class GUI:
 
         cls._init_views()
 
-        cls._view_idx = 0 # take first view as default
         vw = Config.get('startup','view')
         if vw:
             for ii in range(len(GUI._VIEWS)):
                 if GUI._VIEWS[ii].lower() == vw:
                     cls._view_idx = ii
                     break
-        cls._eventbox = Gtk.EventBox()
         cls._box_view_cont.pack_start(cls._eventbox, True, True, 0)
         cls._box_view_cont.reorder_child(cls._eventbox, view_pos)
         cls._eventbox.add(cls.view_widgets[cls._view_idx])
@@ -301,10 +322,9 @@ class GUI:
         # Load clipboard helper library if available, or set to None
         try:
             libclip_file = '{:s}/libpygenda_clipboard.so'.format(ospath.dirname(__file__))
-            cls.lib_clip = ctypes.CDLL(libclip_file)
+            cls._lib_clip = ctypes.CDLL(libclip_file)
         except:
             print('Warning: Failed to load clipboard library', file=stderr)
-            cls.lib_clip = None
 
 
     @classmethod
@@ -350,8 +370,6 @@ class GUI:
     def _init_views(cls) -> None:
         # Get new Gtk Widgets for views.
         # Add view switching options to menu.
-        cls.views = []
-        cls.view_widgets = []
         for v in GUI._VIEWS:
             m = import_module('.pygenda_view_{:s}'.format(v.lower()),package='pygenda')
             cls.views.append(getattr(m, 'View_{:s}'.format(v)))
@@ -554,6 +572,11 @@ class GUI:
         # Callback for new entry signal (menu, softbutton)
         EntryDialogController.newentry()
 
+    @classmethod
+    def event_newtodo(cls, *args) -> None:
+        # Callback for new todo signal (menu, softbutton)
+        TodoDialogController.newtodo()
+
     @classmethod
     def event_edittime(cls, *args) -> None:
         # Callback for change-entry-time signal
@@ -606,7 +629,7 @@ class GUI:
         # Handler to implement "cut" from GUI, e.g. cut clicked in menu
         en = cls.views[cls._view_idx].get_cursor_entry()
         if en and 'SUMMARY' in en:
-            if cls.lib_clip is None:
+            if cls._lib_clip is None:
                 # Don't do fallback - might lead to unexpected data loss
                 print('Warning: No clipboard library, cut not available', file=stderr)
             elif 'RRULE' in en: # repeating entry
@@ -619,7 +642,7 @@ class GUI:
             else:
                 txtbuf = bytes(en['SUMMARY'], 'utf-8')
                 calbuf = en.to_ical()
-                cls.lib_clip.set_cb(ctypes.create_string_buffer(txtbuf),ctypes.create_string_buffer(calbuf))
+                cls._lib_clip.set_cb(ctypes.create_string_buffer(txtbuf),ctypes.create_string_buffer(calbuf))
                 Calendar.delete_entry(en)
                 cls.view_redraw(True)
 
@@ -629,7 +652,7 @@ class GUI:
         # Handler to implement "copy" from GUI, e.g. copy clicked in menu
         en = cls.views[cls._view_idx].get_cursor_entry()
         if en and 'SUMMARY' in en:
-            if cls.lib_clip is None:
+            if cls._lib_clip is None:
                 print('Warning: No clipboard library, fallback to text copy', file=stderr)
                 cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
                 txt = en['SUMMARY']
@@ -637,7 +660,7 @@ class GUI:
             else:
                 txtbuf = bytes(en['SUMMARY'], 'utf-8')
                 calbuf = en.to_ical()
-                cls.lib_clip.set_cb(ctypes.create_string_buffer(txtbuf),ctypes.create_string_buffer(calbuf))
+                cls._lib_clip.set_cb(ctypes.create_string_buffer(txtbuf),ctypes.create_string_buffer(calbuf))
 
 
     @classmethod
@@ -648,9 +671,9 @@ class GUI:
         # First, we try requesting a 'text/calendar' type from the clipboard
         sdat = cb.wait_for_contents(Gdk.Atom.intern('text/calendar', False))
         try:
-            events = iEvent.from_ical(sdat.get_data())
-            ev = events.walk('VEVENT')[0]
-            Calendar.new_entry_from_example(ev, dt_start=cls.cursor_date)
+            ical = iCalendar.from_ical(sdat.get_data())
+            en = ical.walk()[0]
+            cls.views[cls._view_idx].new_entry_from_example(en)
             cls.view_redraw(True)
             return
         except:
@@ -662,8 +685,8 @@ class GUI:
             txt = txt.strip()
             txt = txt.replace('\n',' ')
             txt = txt.replace('\t',' ')
-            # Open a New Entry dialog with description initialised as txt
-            GLib.idle_add(EntryDialogController.newentry,txt)
+            # What type of entry is created will depend on view, so call current view paste fn
+            cls.views[cls._view_idx].paste_text(txt)
 
 
     @classmethod
@@ -678,7 +701,7 @@ class GUI:
             l_template = _('Delete all repeats:\n"{:s}"?')
         else:
             l_template = _('Delete entry:\n"{:s}"?')
-        lab = Gtk.Label(l_template.format(en['SUMMARY'] if 'SUMMARY' in en else 'UNNAMED'))
+        lab = Gtk.Label(l_template.format(en['SUMMARY'] if 'SUMMARY' in en else u' ')) # narrow space
         if (not dialog or not lab): # Sanity check
             raise NameError('Dialog Delete creation failure')
         dialog.set_resizable(False)
@@ -776,8 +799,67 @@ class EntryDialogController:
     TAB_DETAILS = 3
     TAB_COUNT = 4
 
+    dialog = None  # type: Gtk.Dialog
+    wid_desc = None # type: Gtk.Entry
+    _wid_desc_changed_handler = 0
+    wid_date = None # type: WidgetDate
+    wid_tabs = None # type: Gtk.Notebook
+
+    # _empty_desc_allowed has three possible values:
+    #   None - empty desc allowed in dialog (signals delete entry)
+    #   True - empty desc allowed, but can turn False!
+    #   False - empty desc not allowed in dialog (e.g. new entry)
+    _empty_desc_allowed = None # type: Optional[bool]
+
+    wid_timed_buttons = None # type: Gtk.Box
+    # Create widgets for entry time & duration fields
+    wid_time = WidgetTime(dt_time(hour=9))
+    wid_dur = WidgetDuration(timedelta())
+    wid_endtime = WidgetTime(dt_time(hour=9))
+    revs_timedur = None
+    revs_allday = None
+    wid_allday_count = None # type: Gtk.SpinButton
+    _tmdur_handler_time = 0
+    _tmdur_handler_dur = 0
+    _tmdur_handler_endtime = 0
+
+    wid_rep_type = None # type: Gtk.ComboBox
+    revs_repeat = None
+    revs_rep_monthdays = None
+    revs_rep_weekdays = None
+    revs_rep_monthweekdays = None
+    wid_rep_interval = None # type: Gtk.SpinButton
+    wid_rep_forever = None # type: Gtk.CheckButton
+    wid_repbymonthday = None # type: Gtk.ComboBox
+    wid_repbyweekday_day = None # type: Gtk.ComboBox
+    wid_repbyweekday_ord = None # type: Gtk.ComboBox
+    wid_rep_occs = None # type: Gtk.SpinButton
+    wid_rep_enddt = None # type: WidgetDate
+    revs_rep_ends = None
+    rep_occs_determines_end = True
+    _rep_handler_date = 0
+    _rep_handler_time = 0
+    _rep_handler_type = 0
+    _rep_handler_mday = 0
+    _rep_handler_wdayday = 0
+    _rep_handler_wdayord = 0
+    _rep_handler_inter = 0
+    _rep_handler_occs = 0
+    _rep_handler_enddt = 0
+    _lab_rep_exceptions = None # type: Gtk.Label
+    repbymonthday_initialized = False
+    repbyweekday_initialized = False
+    dur_determines_end = False
+    exception_list = [] # type: List[dt_date]
+
+    wid_alarmset = None # type: Gtk.Switch
+
+    wid_status = None # type: Gtk.ComboBox
+    wid_location = None # type: Gtk.Entry
+
+
     @classmethod
-    def init(cls):
+    def init(cls) -> None:
         # Initialiser for singleton class.
         # Called from GUI init_stage2().
 
@@ -801,15 +883,8 @@ class EntryDialogController:
         cls._init_repeatfields()
         cls._init_alarmfields()
         cls._init_detailfields()
-        # Also need to init alarm & detail fields
         cls._init_navigation()
 
-        # cls._empty_desc_allowed has three possible values:
-        #   None - empty desc allowed in dialog (signals delete entry)
-        #   True - empty desc allowed, but can turn False!
-        #   False - empty desc not allowed in dialog (e.g. new entry)
-        cls._empty_desc_allowed = None
-
 
     @classmethod
     def _init_timefields(cls) -> None:
@@ -817,11 +892,7 @@ class EntryDialogController:
         # Called on app startup.
         cls.wid_timed_buttons = GUI._builder.get_object('radiobuttons_istimed').get_children()
 
-        # Create widgets for entry time & duration fields
-        cls.wid_time = WidgetTime(dt_time(hour=9))
-        cls.wid_dur = WidgetDuration(timedelta())
-        cls.wid_endtime = WidgetTime(dt_time(hour=9))
-        # Add to dialog
+        # Add entry time & duration widgets to dialog
         wid_reveal_tm_e = GUI._builder.get_object('revealer_time_e')
         tbox = wid_reveal_tm_e.get_child() # should be a GtkBox
         tbox.add(cls.wid_time)
@@ -871,7 +942,6 @@ class EntryDialogController:
         rbox.show_all()
         cls.revs_rep_ends = cls._revealers_from_ids('revealer_repeat_until_l', 'revealer_repeat_until_e')
 
-        cls.rep_occs_determines_end = True
         cls._rep_handler_date = cls.wid_date.connect('changed', cls._repend_changed,0)
         cls._rep_handler_time = cls.wid_time.connect('changed', cls._repend_changed,0)
         cls._rep_handler_type = cls.wid_rep_type.connect('changed', cls._repend_changed,0)
@@ -1895,3 +1965,76 @@ class ExceptionsDialogController:
     def _sort_func(row1:Gtk.ListBoxRow, row2:Gtk.ListBoxRow) -> int:
         # Used as sort function for listbox.
         return row1.get_child().compare(row2.get_child())
+
+
+# Singleton class to manage Todo dialog
+class TodoDialogController:
+    dialog = None # type: Gtk.Dialog
+    wid_desc = None # type: Gtk.Entry
+
+    @classmethod
+    def init(cls) -> None:
+        # Initialiser for singleton class.
+        # Called from GUI init_stage2().
+
+        # Get some references to dialog elements in glade
+        cls.dialog = GUI._builder.get_object('dialog_todo')
+        if (not cls.dialog): # Sanity check
+            raise NameError('Dialog Todo not found')
+
+        cls.wid_desc = GUI._builder.get_object('todo_desc')
+
+
+    @classmethod
+    def newtodo(cls, txt:str=None) -> None:
+        # Called to implement "new todo" from GUI, e.g. button
+        cls.dialog.set_title(_('New To-do'))
+        response,ei = cls._do_todo_dialog(txt=txt)
+        if response==Gtk.ResponseType.OK and ei.desc:
+            Calendar.new_entry(ei)
+            GUI.view_redraw(True)
+
+
+    @classmethod
+    def edittodo(cls, en:iTodo) -> None:
+        # Called to implement "edit todo" from GUI
+        cls.dialog.set_title(_('Edit To-do'))
+        response,ei = cls._do_todo_dialog(entry=en)
+        if response==Gtk.ResponseType.OK:
+            if ei.desc:
+                Calendar.update_entry(en, ei)
+                GUI.view_redraw(True)
+            else: # Description text has been deleted in dialog
+                GUI.dialog_deleteentry(en)
+
+
+    @classmethod
+    def _do_todo_dialog(cls, txt:str=None, entry:iTodo=None) -> Tuple[int,EntryInfo]:
+        # Do the core work displaying todo dialog and extracting result.
+        if entry is not None:
+            # existing entry - take values
+            cls.wid_desc.set_text(entry['SUMMARY'] if 'SUMMARY' in entry else'')
+            cls.wid_desc.grab_focus()
+        elif txt is None:
+            cls.wid_desc.set_text('') # clear text
+        else:
+            cls.wid_desc.set_text(txt)
+            cls.wid_desc.set_position(len(txt))
+        cls.wid_desc.grab_focus_without_selecting()
+
+        try:
+            while True:
+                response = cls.dialog.run()
+                break
+        finally:
+            cls.dialog.hide()
+
+        return response,cls._get_entryinfo()
+
+
+    @classmethod
+    def _get_entryinfo(cls) -> EntryInfo:
+        # Decipher entry fields and return info as an EntryInfo object.
+        desc = cls.wid_desc.get_text()
+        ei = EntryInfo(type=EntryInfo.TYPE_TODO, desc=desc)
+        return ei
